
#ifdef WINDOWS
	#define _UNICODE
	#define UNICODE
	#ifndef _MSC_VER
		//#define WINVER 0x0501
		//#define _WIN32_IE 0x0400
		//#define __MSVCRT_VERSION__ 0x0700
	#endif
	#include <windows.h>
	typedef WCHAR wchar;
	#ifdef _MSC_VER
		#define ENTRY
		typedef unsigned __int64 uint64_t;
		typedef __int64 int64_t;
		//BOOL WINAPI GetFileSizeEx(HANDLE hFile, PLARGE_INTEGER lpFileSize);
		#ifndef INVALID_FILE_ATTRIBUTES
			#define INVALID_FILE_ATTRIBUTES	((DWORD)-1)
		#endif
		#ifndef INVALID_SET_FILE_POINTER
			#define INVALID_SET_FILE_POINTER	((DWORD)-1)
		#endif
	#else
		#include <stdint.h>
		#include <stddef.h>
		#define ENTRY
	#endif
#else
	#include <stdint.h>
	#include <stddef.h>
	#include <iconv.h>
	#define ENTRY
#endif

//[of]:Error
typedef enum
{
	OK,	// No error
	ERR_INVALID_UTF8,	// Invalid UTF-8 encoded data
	ERR_NUL_CODE_UNIT,	// The converted data contains a NUL code unit
	ERR_SYSTEM	// System error (errno on Unix, GetLastError on Windows)
} SysErrorCode;
//[cf]
//[of]:File
typedef enum
{
	OPEN_READ,
	OPEN_WRITE,
} OpenMode;

typedef enum
{
	SM_SET,
	SM_CURRENT,
	SM_END,
} SeekMode;
//[cf]
//[of]:Debug
ENTRY void csys_trap();
//[cf]
//[of]:Allocator
// Memory statistics
// These global variables are public, so any program can report a leak
struct MemoryStatistics
{
	size_t totalBlock;
	size_t leftBlock;
	size_t leftBytes;
	size_t maxBytes;
};

ENTRY void* csys_allocate (size_t size);
ENTRY void csys_free (void* p, size_t size);
//[cf]
//[of]:Text
#ifdef WINDOWS
	typedef wchar CodeUnit;
#else
	typedef char CodeUnit;
#endif

typedef CodeUnit* String;

//[of]:MemoryBuffer
// The structure is identical to the Copper counterpart, so a MemoryBuffer
// can be passed to a C function.
typedef struct
{
	char*	base;
	size_t	size;
	size_t	allocated;
} MemoryBuffer;

void MemoryBuffer_initialize(MemoryBuffer* sb);
void MemoryBuffer_release(MemoryBuffer* sb);
void MemoryBuffer_add(MemoryBuffer* sb, const char* data, size_t size);
char* MemoryBuffer_reserve(MemoryBuffer* sb, size_t len);
//[cf]
//[of]:StringBuffer
#define STRING_BUFFER_LOCAL_SIZE 128

// The structure is identical to the Copper counterpart, so a StringBuffer
// can be passed to a C function.
typedef struct
{
	CodeUnit*	base;
	int	size;
	int	allocated;
	CodeUnit	local[STRING_BUFFER_LOCAL_SIZE];
} StringBuffer;

void StringBuffer_initialize(StringBuffer* sb);
void StringBuffer_release(StringBuffer* sb);
void StringBuffer_addString(StringBuffer* sb, const String str);
String StringBuffer_newString(StringBuffer* sb);
String StringBuffer_reserve(StringBuffer* sb, size_t len);

#ifdef WINDOWS
	SysErrorCode StringBuffer_addWide(StringBuffer* sb, const wchar* str);
#else
	SysErrorCode StringBuffer_addLocale(StringBuffer* sb, const char* str);
#endif
//[cf]
//[cf]
//[of]:Encoding
ENTRY void csys_deleteString(String string);

#ifdef WINDOWS
	SysErrorCode csys_newStringFromWide(String* out, const wchar* input);
	SysErrorCode csys_newWideFromString(wchar**out, String input);
	void csys_deleteWide(const wchar* string);
#else
	SysErrorCode csys_newStringFromLocale(String* out, const char* input);
	SysErrorCode csys_newCStringFromLocale(String* out, const char* input);
	SysErrorCode csys_newLocaleFromString(char**out, String input);
	void csys_deleteLocale(const char* string);
#endif

// The longest sequence in bytes for any encoding
// * UTF-16: 2 16-bit units (low and high surrogates)
// * UTF-8: 4 8-bit units
// * Other: mostly single byte encodings
//
#define MAX_SEQUENCE_SIZE 4

#ifdef WINDOWS
	typedef enum 
	{
		CE_UTF16LE,
		CE_UTF16BE,
		CE_OTHER
	} ConverterEncoding;
#endif

//[of]:TextDecoder
typedef struct 
{
	#ifdef WINDOWS
		ConverterEncoding	fromEncoding;
		UINT	fromCodePage;
		MemoryBuffer	buffer; // All the input is stored before being converted as Windows API does not help with streaming
	#else
		iconv_t	cd;
		size_t 	minibufSize;	// Number of bytes in the mini-buffer
		char	minibuf[MAX_SEQUENCE_SIZE];
	#endif
} TextDecoder;

ENTRY SysErrorCode csys_TextDecoder_open(const char* fromCode, TextDecoder** decoder);
ENTRY SysErrorCode csys_TextDecoder_convert(TextDecoder* decoder, StringBuffer* sb, const char* input, size_t size);
ENTRY SysErrorCode csys_TextDecoder_close(TextDecoder* decoder);
//[cf]
//[of]:TextEncoder
typedef struct 
{
	#ifdef WINDOWS
		ConverterEncoding	toEncoding;
		UINT	toCodePage;
	#else
		iconv_t	cd;
	#endif
} TextEncoder;

ENTRY SysErrorCode csys_TextEncoder_open(const char* toCode, TextEncoder** encoder);
ENTRY SysErrorCode csys_TextEncoder_convert(TextEncoder* encoder, MemoryBuffer* mb, String input, int size);
ENTRY SysErrorCode csys_TextEncoder_close(TextEncoder* encoder);
//[cf]
//[cf]
//[of]:Time
#ifdef WINDOWS
	uint64_t csys_windowsTimeToTime(uint64_t time);
#else
	uint64_t csys_unixTimeToTime(uint64_t time);
#endif
//[cf]
